#include <Arduino.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#undef WHITE
#undef BLACK
#include <SparkFunTMP102.h>
#include <Adafruit_NeoPixel.h>

#include "touch.h"

enum Pinout {
  DRIVE2 = 0,
  DRIVE1 = 1,
  LED_W = 2,
  WS2812_DISPLAY = 2,
  LED_Y = 3,
  EN_MEASURE_CURRENT = 3,
  LED_B = 4,
  LED_G = 5,
  LED_R = 6,
  MATRIX_IN1 = 7,
  SBU1 = 8,
  SBU2 = 9,
  VBUS_DET = 10,
  BOOT2 = 11,
  I2C_OLED_SDA = 12,
  I2C_OLED_SCL = 13,
  WS2811 = 14,
  TX_EN = 15,
  TX = 16,
  RX = 17,
  MATRIX_IN2 = 18,
  REED3 = 19,
  REED4 = 20,
  REED2 = 21,
  REED1 = 22,
  MATRIX_IN3 = 23,
  DIGOUT1 = 24,
  DIGOUT2 = 25,
  ANALOG_IN1 = 26,
  CURRENT2 = 27,
  CURRENT1 = 28,
  MEASURE_VCC = 29,
};

const int SCREEN_ADDRESS = 0x3C;
Adafruit_SSD1306 display(128, 64, &Wire, -1);
bool hasDisplay;

#include "logo.h"

TMP102 sensor0, sensor1;

HeizungTouch touch;
Adafruit_NeoPixel pixelsDisplay(60, WS2812_DISPLAY, NEO_GRB + NEO_KHZ800);

static bool initial = true;

void setup() {
  rp2040.enableDoubleResetBootloader();

  pinMode(DIGOUT1, OUTPUT);
  digitalWrite(DIGOUT1, LOW);

  Serial.begin(9600);  // USB serial, baudrate is ignored
  if (0) {
    while (!Serial)
      ;
    Serial.println("a1");
  }

  pinMode(DRIVE1, OUTPUT); digitalWrite(DRIVE1, LOW);
  pinMode(DRIVE2, OUTPUT); digitalWrite(DRIVE2, LOW);
  pinMode(WS2812_DISPLAY, OUTPUT);
  pinMode(3, OUTPUT);  digitalWrite(EN_MEASURE_CURRENT, LOW);
  pinMode(4, OUTPUT);
  pinMode(5, OUTPUT);
  pinMode(6, OUTPUT);

  pinMode(BOOT2, INPUT_PULLUP);
  pinMode(WS2811, INPUT_PULLUP);
  pinMode(REED1, OUTPUT);
  pinMode(REED2, OUTPUT);
  pinMode(REED3, OUTPUT);
  pinMode(REED4, OUTPUT);
  pinMode(MATRIX_IN1, INPUT_PULLUP);
  pinMode(MATRIX_IN3, INPUT_PULLUP);

  pinMode(TX_EN, OUTPUT);
  digitalWrite(TX_EN, LOW);

  pixelsDisplay.begin();
  pixelsDisplay.clear();
  pixelsDisplay.show();

  Serial1.setTX(TX);  // UART0
  Serial1.setRX(RX);
  Wire.setSDA(I2C_OLED_SDA);  // I2C0
  Wire.setSCL(I2C_OLED_SCL);

  Serial1.begin(115200);
  Wire.begin();
  Wire.setClock(100000);

  display.setRotation(2);
  hasDisplay = display.begin(SSD1306_SWITCHCAPVCC, SCREEN_ADDRESS);
  if (!hasDisplay) {
    Serial.println(F("SSD1306 allocation failed"));
  } else {
    display.clearDisplay();
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);
    display.display();
  }

  if (sensor0.begin(0x48)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor0.setConversionRate(2);
    sensor0.setExtendedMode(1);
    sensor0.wakeup();
  }
  if (sensor1.begin(0x49)) {
    //0-3: 0:0.25Hz, 1:1Hz, 2:4Hz, 3:8Hz
    sensor1.setConversionRate(2);
    sensor1.setExtendedMode(1);
    sensor1.wakeup();
  }

  touch.init();
}

void displayPrintHexFixed(uint32_t x, int width) {
  while (width > 1 && (x >> (4*(width-1))) == 0) {
    display.write('0');
    width--;
  }
  display.print(x, HEX);
}

void readButtons(bool buttons[4]) {
  // Button Matrix:
  // 1: SDA, IN1
  // 2: SDA, IN3
  // 3: SCL, IN1
  // 4: SCL, IN4
  Wire.end();
  pinMode(I2C_OLED_SDA, OUTPUT);
  pinMode(I2C_OLED_SCL, INPUT_PULLUP);
  delayMicroseconds(1);
  buttons[0] = !digitalRead(MATRIX_IN1);
  buttons[1] = !digitalRead(MATRIX_IN3);
  pinMode(I2C_OLED_SDA, INPUT_PULLUP);
  pinMode(I2C_OLED_SCL, OUTPUT);
  delayMicroseconds(1);
  buttons[2] = !digitalRead(MATRIX_IN1);
  buttons[3] = !digitalRead(MATRIX_IN3);
  pinMode(I2C_OLED_SDA, INPUT_PULLUP);
  pinMode(I2C_OLED_SCL, INPUT_PULLUP);
  Wire.begin();
}

// show remaining time in milliseconds and use short timeout for 3rd button
constexpr bool DEBUG_TIME = false;
// show touch bits on display
constexpr bool DEBUG_TOUCH = false;
// make LEDs a lot darker so it is easier to take a photo
constexpr bool LEDS_FOR_PHOTO = false;

typedef uint16_t temperature_t;
constexpr temperature_t TEMPERATURE_MIN = 100;  // 10.0 deg C
constexpr temperature_t TEMPERATURE_MAX = 300;
constexpr temperature_t TEMPERATURE_STANDBY = 150;
typedef unsigned long millis_t;
typedef signed long millis_diff_t;
constexpr millis_t MINUTES = 60*1000;
constexpr millis_t HOURS = 3600*1000;
constexpr millis_t TIME_UNTIL_STANDBY_DEFAULT = 2*HOURS + MINUTES/2;
constexpr millis_t TIME_UNTIL_STANDBY_DEFAULT2 = DEBUG_TIME ? 10*1000 : TIME_UNTIL_STANDBY_DEFAULT;
constexpr millis_t TIME_UNTIL_STANDBY_MAX = 48*HOURS - MINUTES/2;
constexpr millis_t TIME_DISABLED = 0;

typedef uint32_t color_t;

color_t blend(color_t x, color_t y, uint8_t alpha) {
  // Adafruit Neopixel library combines the values into a uint32_t. We extract them, again.
  uint16_t wx = (x >> 24) & 0xff, wy = (y >> 24) & 0xff;
  uint16_t rx = (x >> 16) & 0xff, ry = (y >> 16) & 0xff;
  uint16_t gx = (x >>  8) & 0xff, gy = (y >>  8) & 0xff;
  uint16_t bx = (x >>  0) & 0xff, by = (y >>  0) & 0xff;

  uint8_t w = (wx * (255-alpha) + wy * alpha) / 255;
  uint8_t r = (rx * (255-alpha) + ry * alpha) / 255;
  uint8_t g = (gx * (255-alpha) + gy * alpha) / 255;
  uint8_t b = (bx * (255-alpha) + by * alpha) / 255;

  return Adafruit_NeoPixel::Color(r, g, b, w);
}

// clamp alpha to 0..255. In other words:
// alpha <= 0: take x
// alpha 1..254: blend
// alpha >= 255: take y
color_t blendWithSaturatingAlpha(color_t x, color_t y, int alpha) {
  return blend(x, y, min(255, max(0, alpha)));
}

uint8_t led_number_to_thermo_ring_index(uint8_t led) {
  // actual ring starts on the right and goes counter-clockwise ("mathematical angle")
  // but we want from lower left, clockwise to lower right
  int j;
  if (led < 38)
    return 37-led;
  else if (led < 38+15)
    return 100;  // lower part
  else
    return 38 + (59-led);
}
constexpr uint8_t thermo_ring_index_count = 60-16;

void loop() {
  static uint16_t wantedTemperature = TEMPERATURE_STANDBY;
  static long wantedTemperatureUntil = TIME_DISABLED;
  bool changed = false;
  static uint8_t double_tap_fine_angle = 255;

  // Calculate diff with unsigned numbers because signed overflow is undefined behaviour
  // but convert the result to signed so we can compare it to zero (rather than LONG_MAX/2).
  const millis_t now = millis();
  const millis_diff_t time_to_standby = (millis_diff_t)(wantedTemperatureUntil - now);
  if (time_to_standby <= 0) {
    wantedTemperatureUntil = 0;
    wantedTemperature = TEMPERATURE_STANDBY;
  }

  static int button_timer = 0;
  bool buttons[4];
  if (button_timer < 500)
    button_timer++;
  else {
    readButtons(buttons);

    if (buttons[0] && wantedTemperature > TEMPERATURE_MIN) {
      wantedTemperature--;
      changed = true;
    }
    if (buttons[1] && wantedTemperature < TEMPERATURE_MAX) {
      wantedTemperature++;
      changed = true;
      if (wantedTemperature > TEMPERATURE_STANDBY && (wantedTemperatureUntil == 0 || time_to_standby < TIME_UNTIL_STANDBY_DEFAULT)) {
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT;
        if (wantedTemperatureUntil == TIME_DISABLED)
          wantedTemperatureUntil++;  // avoid special value
      }
    }
    if (buttons[2] && wantedTemperature > TEMPERATURE_STANDBY) {
      if (0)
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT;
      else
        // very small timeout for debugging
        wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT2;

      if (wantedTemperatureUntil == TIME_DISABLED)
        wantedTemperatureUntil++;  // avoid special value
    }
    if (buttons[3] && wantedTemperatureUntil != TIME_DISABLED) {
      changed = true;
      if (time_to_standby >= TIME_UNTIL_STANDBY_MAX)
        ;
      else if (time_to_standby < 1*HOURS)
        wantedTemperatureUntil += 1 * MINUTES;
      else if (time_to_standby < 2*HOURS)
        wantedTemperatureUntil += 2 * MINUTES;
      else if (time_to_standby < 10*HOURS)
        wantedTemperatureUntil += 5 * MINUTES;
      else
        wantedTemperatureUntil += 20 * MINUTES;

      if (wantedTemperatureUntil == TIME_DISABLED)
        wantedTemperatureUntil++;  // avoid special value
    }
  }

  touch.update(now);
  if (DEBUG_TOUCH || 1)
    changed |= touch.hasChanged();
  if (1) {
    //FIXME filter too short events, e.g. skip this whole block for some holdoff time after an event and
    //   only accept those that are still present after that time
    //FIXME we don't have to track double-tap for each area, I think (if we have hold-off)
    auto touched = touch.touchedBits();
    auto changed = touch.changedBits();
    static millis_t touch_event_time = 0;
    static std::bitset<32> touch_event_press = 0, touch_event_tap = 0,
      touch_event_double_press = 0, touch_event_double_tap = 0;
    int double_tap_index = -1;
    if (changed.any()) {
      for (int i=0; i<32; i++) {
        using enum TouchAreaType;
        auto type = TOUCH_AREAS[i].type;
        if (changed[i] && (type == SPOKE || type == SKREW)) {
          touch_event_time = now;
          bool pressed = touched[i];
          if (pressed && touch_event_tap[i]) {
            touch_event_double_press[i] = true;
          } else if (pressed) {
            touch_event_press[i] = true;
          } else if (touch_event_double_press[i]) {
            touch_event_double_tap[i] = true;
            double_tap_index = i;
          } else if (touch_event_press[i]) {
            touch_event_tap[i] = true;
          } else {
            // released but touch_event_press is not set,
            // i.e. long press or touch_event_press has been
            // reset while handling a double tap
          }
        }
      }

      if (double_tap_index >= 0 && TOUCH_AREAS[double_tap_index].type == TouchAreaType::SPOKE) {
        auto angle = TOUCH_AREAS[double_tap_index].angle;
        auto fine_angle = 4*angle + 59;
        for (int i=0; i<32; i++) {
          // look at touch_event_press for other areas that have been pressed at a similar time
          // but either only once or are still pressed
          if (i != double_tap_index && touch_event_press[i] && TOUCH_AREAS[i].type == TouchAreaType::SPOKE) {
            auto angle2 = TOUCH_AREAS[i].angle;
            if ((angle2+1)%15 == angle)
              fine_angle -= 2;
            else if (angle2 == (angle+1)%15)
              fine_angle += 2;
          }
        }

        fine_angle %= 60;
        double_tap_fine_angle = fine_angle;

        auto idx = led_number_to_thermo_ring_index(fine_angle);
        auto temperature = TEMPERATURE_MIN + (TEMPERATURE_MAX - TEMPERATURE_MIN) * idx / thermo_ring_index_count;
        wantedTemperature = temperature;
        changed = true;
        if (wantedTemperature > TEMPERATURE_STANDBY && (wantedTemperatureUntil == 0 || time_to_standby < TIME_UNTIL_STANDBY_DEFAULT)) {
          wantedTemperatureUntil = now + TIME_UNTIL_STANDBY_DEFAULT;
          if (wantedTemperatureUntil == TIME_DISABLED)
            wantedTemperatureUntil++;  // avoid special value
        }

        touch_event_press = 0;
        touch_event_tap = 0;
        touch_event_double_press = 0;
        touch_event_double_tap = 0;
      }
    } else if ((millis_diff_t)(now - touch_event_time) > 300) {
      touch_event_press = 0;
      touch_event_tap = 0;
      touch_event_double_press = 0;
      touch_event_double_tap = 0;
    }
  }

  auto temp0 = sensor0.readTempC();
  auto temp1 = sensor1.readTempC();
  auto temp = !isnan(temp1) ? temp1 : temp0;
  auto actualTemperature = (temperature_t)(temp * 10);

  // very stupid control loop but should be good enough, for now
  // (Good control loop is not part of minimum viable product and can later be done in Python if anyone cares to do it.)
  bool heatOn = (actualTemperature < wantedTemperature);
  digitalWrite(DRIVE1, heatOn);
  digitalWrite(DRIVE2, heatOn);

  static int display_timer = 0;
  if (display_timer < 500 && !changed)
    display_timer++;
  else {
    display_timer = 0;
  }

  if (DEBUG_TOUCH) {
    if (changed || initial || display_timer == 0) {
      pixelsDisplay.clear();

      uint32_t touched = (touch.touched[1] << 16) | touch.touched[0];
      uint32_t touchedPrev = (touch.touchedPrev[1] << 16) | touch.touchedPrev[0];
      bool unexpectedTouch = false;
      for (int i=0; i<32; i++) {
        using enum TouchAreaType;

        if (!(touched & (1<<i)))
          continue;
        bool isNew = !(touchedPrev & (1<<i));

        switch (TOUCH_AREAS[i].type) {
          case NONE:
            unexpectedTouch = true;
            break;
          case SPOKE:
            pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle*(60/15) + 59) % 60,
              isNew ? Adafruit_NeoPixel::Color(32, 0, 0) : Adafruit_NeoPixel::Color(8, 8, 0));
            break;
          case SKREW:
            pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle*(60/15) + 59) % 60,
              isNew ? Adafruit_NeoPixel::Color(0, 0, 32) : Adafruit_NeoPixel::Color(0, 8, 8));
            break;
          case INNER:
            for (int j=0; j<8; j++) {
              pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle+j)*(60/15) + 0,
                Adafruit_NeoPixel::Color(0, 2, 0));
            }
            break;
          case MIDDLE:
            for (int j=0; j<8; j++) {
              pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle+j)*(60/15) + 1,
                Adafruit_NeoPixel::Color(0, 2, 0));
            }
            break;
          case OUTER:
            for (int j=0; j<8; j++) {
              pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle+j)*(60/15) + 2,
                Adafruit_NeoPixel::Color(0, 2, 0));
            }
            break;
          default:
            break;
        }

        if (unexpectedTouch)
          pixelsDisplay.setPixelColor(0, Adafruit_NeoPixel::Color(32, 0, 0));
      }

      pixelsDisplay.show();
    }
  } else if (changed || initial || display_timer == 0) {
    initial = false;

    auto BLACK  = pixelsDisplay.Color(0, 0, 0);
    auto WHITE  = pixelsDisplay.Color(8, 8, 8);
    auto BLUE   = pixelsDisplay.Color(0, 0, 24);
    auto GREEN  = pixelsDisplay.Color(0, 8, 0);
    auto ORANGE = pixelsDisplay.Color(24, 16, 0);

    if (LEDS_FOR_PHOTO && 0) {
      WHITE  = pixelsDisplay.Color(3, 3, 2);
      ORANGE = pixelsDisplay.Color(12, 6, 0);
    } else if (LEDS_FOR_PHOTO) {
      WHITE  = pixelsDisplay.Color(8, 8, 4);
    }

    int pixelsOnWanted = 256 + (wantedTemperature - TEMPERATURE_MIN) * 256 * thermo_ring_index_count / (TEMPERATURE_MAX - TEMPERATURE_MIN);
    int pixelsOnActual = 256 + (actualTemperature - TEMPERATURE_MIN) * 256 * thermo_ring_index_count / (TEMPERATURE_MAX - TEMPERATURE_MIN);
    for (int i=0; i<pixelsDisplay.numPixels(); i++) {
      // actual ring starts on the right and goes counter-clockwise ("mathematical angle")
      // but we want from lower left, clockwise to lower right
      int j = led_number_to_thermo_ring_index(i);

      bool wanted = (j < pixelsOnWanted/256);
      int actual = (pixelsOnActual - 256*j);
      auto color = BLACK;
      if (j == 100) {
        color = BLACK;  // lower part of the ring
      } else if (wanted) {
        // actual > 0:  actual and wanted       -> WHITE
        // actual <= 0: too cold, needs heating -> ORANGE
        color = blendWithSaturatingAlpha(ORANGE, WHITE, actual);
      } else {
        if (wantedTemperatureUntil == TIME_DISABLED)
          // maybe too hot but for standby temperature
          color = blendWithSaturatingAlpha(BLACK, GREEN, actual);
        else
          // maybe too hot -> needs cooling
          color = blendWithSaturatingAlpha(BLACK, BLUE, actual);
      }

      pixelsDisplay.setPixelColor(i, color);
    }

    if (1) {
      // We have our own copy different from touch.touchedPrev because this code
      // might not run every cycle (although it does, at the moment).
      static uint32_t touchedPrev = 0;
      uint32_t touched = (touch.touched[1] << 16) | touch.touched[0];
      uint32_t touchedNew = touched & ~touchedPrev;
      touchedPrev = touched;
      for (int i=0; i<32; i++) {
        using enum TouchAreaType;

        auto type = TOUCH_AREAS[i].type;
        if ((touchedNew & (1<<i)) && (type == SPOKE || type == SKREW)) {
          pixelsDisplay.setPixelColor((TOUCH_AREAS[i].angle*(60/15) + 59) % 60,
            type == SPOKE ? Adafruit_NeoPixel::Color(32, 0, 0) : Adafruit_NeoPixel::Color(0, 0, 32));
        }
      }

      if (double_tap_fine_angle != 255 && 0) {
        pixelsDisplay.setPixelColor(double_tap_fine_angle,
          Adafruit_NeoPixel::Color(0, 32, 0));
      }
    }

    pixelsDisplay.show();
  }

  if (changed || initial || display_timer == 0) {
    display_timer = 0;

    display.fillRect(0, 0, 128, 64, SSD1306_BLACK);
    display.drawBitmap(0, 0, logo_data, logo_width, logo_height, SSD1306_WHITE);

    display.setTextSize(1);
    display.setTextColor(SSD1306_WHITE);
    display.setCursor(0, 1);
    display.print(F("Soll: "));
    display.print(wantedTemperature/10);
    display.print(F("."));
    display.print(wantedTemperature%10);
    display.println();
    display.print(F("Ist:  "));
    display.print(actualTemperature/10);
    display.print(F("."));
    display.print(actualTemperature%10);
    display.println();

    if (wantedTemperatureUntil != TIME_DISABLED) {
      display.print(F("noch: "));
      int x = time_to_standby/1000/60/60;
      display.print(x);
      display.print(F(":"));
      x = time_to_standby/1000/60 % 60;
      if (x < 10)
        display.print("0");
      display.println(x);
    } else {
      display.println(F("noch: ----"));
    }

    display.print(F("int:  "));
    display.println(temp0, 1);

    //FIXME show as bar
    display.print(F("Heat: "));
    display.println(heatOn);

    //FIXME measure and display resistance of heating element (and whether there is one)

    if (DEBUG_TOUCH) {
      display.print(F("T:"));
      if (touch.available[0])
        displayPrintHexFixed(touch.touched[1], 4);
      else
        display.write("----");
      if (touch.available[1])
        displayPrintHexFixed(touch.touched[0], 4);
      else
        display.write("----");
      display.println();
    }

    if (DEBUG_TIME) {
      if (wantedTemperatureUntil == TIME_DISABLED)
        display.print(F("no timeout"));
      else
        display.print(time_to_standby);
      //display.println(buttons[3] ? ", X" : ", _");
      display.println();
    }

    display.setCursor(3, 57);
    display.println(F("---   +++   2h    +t"));

    display.display();
  }

  if (!digitalRead(BOOT2)) {
    analogWrite(LED_R, 128);
  } else {
    analogWrite(LED_R, 0);
  }

  delay(1);
}